package settings;

import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.List;

import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;

import android.content.Context;
import android.text.InputType;
import android.util.Log;
import android.view.Gravity;
import android.view.LayoutInflater;
import android.view.MotionEvent;
import android.view.View;
import android.view.ViewGroup;
import android.widget.BaseAdapter;
import android.widget.CheckBox;
import android.widget.EditText;
import android.widget.LinearLayout;
import android.widget.Spinner;
import android.widget.TextView;
import rs.corelibrary.R;

@SuppressWarnings({ "rawtypes", "unchecked" })
public class AppSettings {

	private static final String VALUE_TAG = "value";
	private static final String TYPE_TAG = "type";
	private static final String DESCRIPTION_TAG = "description";
	private static final String OPTIONS_TAG = "options";

	// private static Map<Class<?>, List<Field>> CACHE = new HashMap<>();

	public static boolean loadFromJson(Object o, String s) {
		if (o == null || s == null)
			return false;
		try {
			return loadFromJson(o, new JSONObject(s));
		} catch (JSONException e) {
			return false;
		}
	}

	public static boolean loadFromJson(Object o, JSONObject s) {
		if (o == null || s == null)
			return false;
		Class<?> c = o.getClass();
		while (c != Object.class) {
			for (Field f : c.getDeclaredFields()) {
				if (f.isAnnotationPresent(SettingsField.class)) {
					Class<?> fType = f.getType();
					SettingsField annotation = f.getAnnotation(SettingsField.class);
					if (s.has(annotation.name()))
						try {
							JSONObject val = s.getJSONObject(annotation.name());
							if (val.has(VALUE_TAG)) {
								f.setAccessible(true);
								if (fType == Integer.class || fType == int.class)
									f.setInt(o, val.getInt(VALUE_TAG));
								else if (fType == Long.class || fType == long.class)
									f.setLong(o, val.getLong(VALUE_TAG));
								else if (fType == Boolean.class || fType == boolean.class)
									f.setBoolean(o, val.getBoolean(VALUE_TAG));
								else if (String.class.isAssignableFrom(fType))
									f.set(o, val.getString(VALUE_TAG));
								else if (List.class.isAssignableFrom(fType)) {
									List l = (List) f.get(o);
									JSONArray a = val.getJSONArray(VALUE_TAG);
									for (int i = 0; i < a.length(); i++) {
										JSONObject elem = a.getJSONObject(i);
										String eType = annotation.type();
										if ("int".equals(eType))
											l.add(elem.getInt(VALUE_TAG));
										else if ("long".equals(eType))
											l.add(elem.getLong(VALUE_TAG));
										else if ("bool".equals(eType))
											l.add(elem.getBoolean(VALUE_TAG));
										else if ("byte".equals(eType))
											l.add((byte) (elem.getInt(VALUE_TAG) & 0xFF));
										else if ("string".equals(eType))
											l.add(elem.getString(VALUE_TAG));
										else {
											try {
												Object eObject = annotation.elementClass().newInstance();
												loadFromJson(eObject, elem);
												l.add(eObject);
											} catch (Exception e) {
												Log.e("RSCore", "Instance error", e);
											}
										}
									}
								} else if (Object.class.isAssignableFrom(fType))
									loadFromJson(f.get(o), val.getJSONObject(VALUE_TAG));
							}
						} catch (Exception e) {
							Log.e("RSCore", "Load error " + c.getName() + "." + annotation.name(), e);
						}
				}
			}
			c = c.getSuperclass();
		}
		return true;
	}

	private static JSONObject setValue(Object o, SettingsField annotation, Context ctx) throws JSONException {
		JSONObject result = new JSONObject();
		if (o == null)
			return result;
		result.put(OPTIONS_TAG, annotation.options());
		if (annotation.descrId() != 0)
			result.put(DESCRIPTION_TAG, ctx.getString(annotation.descrId()));
		else
			result.put(DESCRIPTION_TAG, annotation.description());
		Class<?> c = o.getClass();
		if (c == Boolean.class || c == boolean.class) {
			result.put(VALUE_TAG, ((Boolean) o).booleanValue());
			result.put(TYPE_TAG, "bool");
		} else if (c == Integer.class || c == int.class) {
			result.put(VALUE_TAG, (int) ((Number) o).intValue());
			result.put(TYPE_TAG, "int");
		} else if (c == Long.class || c == long.class) {
			result.put(VALUE_TAG, (int) ((Number) o).longValue());
			result.put(TYPE_TAG, "long");
		} else if (c == Byte.class || c == byte.class) {
			result.put(VALUE_TAG, (int) ((Number) o).byteValue());
			result.put(TYPE_TAG, "byte");
		} else if (String.class.isAssignableFrom(c)) {
			result.put(VALUE_TAG, o.toString());
			if (annotation.type().isEmpty())
				result.put(TYPE_TAG, "string");
			else
				result.put(TYPE_TAG, annotation.type());
		} else if (List.class.isAssignableFrom(c)) {
			result.put(TYPE_TAG, "list");
			JSONArray a = new JSONArray();
			for (Object o1 : (List) o)
				a.put(store(o1, ctx));
			result.put(VALUE_TAG, a);
		} else if (Object.class.isAssignableFrom(c)) {
			result.put(VALUE_TAG, store(o, ctx));
			result.put(TYPE_TAG, c.getName());
		}
		return result;
	}

	public static JSONObject store(Object o, Context ctx) {
		JSONObject result = new JSONObject();
		if (o != null) {
			Class<?> c = o.getClass();
			while (c != Object.class) {
				for (Field f : c.getDeclaredFields()) {
					if (f.isAnnotationPresent(SettingsField.class))
						try {
							f.setAccessible(true);
							SettingsField annotation = f.getAnnotation(SettingsField.class);
							result.put(annotation.name(), setValue(f.get(o), annotation, ctx));
						} catch (Exception e) {
							Log.e("RSCore", "Settings error ", e);
						}
				}
				c = c.getSuperclass();
			}
		}

		return result;
	}

	private static class SelectionAdapter extends BaseAdapter {

		private JSONArray data;
		private LayoutInflater inflater;

		public SelectionAdapter(Context ctx, JSONArray data) {
			this.data = data;
			inflater = LayoutInflater.from(ctx);
		}

		@Override
		public int getCount() {
			return data.length();
		}

		@Override
		public JSONObject getItem(int position) {
			try {
				return data.getJSONObject(position);
			} catch (JSONException jse) {
				return null;
			}
		}

		@Override
		public long getItemId(int position) {
			try {
				return getItem(position).getLong("k");
			} catch (JSONException jse) {
				return position;
			}
		}

		public int positionById(long id) {
			for (int i = 0; i < getCount(); i++)
				if (getItemId(i) == id)
					return i;
			return -1;
		}

		@Override
		public View getView(int position, View v, ViewGroup vg) {
			if (v == null)
				v = inflater.inflate(android.R.layout.simple_spinner_item, vg, false);
			try {
				((TextView) v).setText(getItem(position).getString("v"));
			} catch (JSONException jse) {
			}
			return v;
		}

	}

	public static void bind(Object o, final ViewGroup container) {
		bind(o, container, 0);
	}

	public static boolean store(ViewGroup vg) {
		for (int i = 0; i < vg.getChildCount(); i++) {
			View v = vg.getChildAt(i);
			if (v.getTag() instanceof EditorInfo) {
				EditorInfo info = (EditorInfo) v.getTag();
				Class<?> fType = info.f.getType();
				View editor = ((ViewGroup) v).getChildAt(1);
				try {
					if (Integer.class == fType || int.class == fType) {
						if (editor instanceof Spinner) {
							info.f.setInt(info.o, (int) ((Spinner) editor).getSelectedItemId());
						} else
							info.f.setInt(info.o, Integer.parseInt(((TextView) editor).getText().toString()));
					} else if (Boolean.class == fType || boolean.class == fType) {
						editor = ((ViewGroup) v).getChildAt(0);
						info.f.setBoolean(info.o, ((CheckBox) editor).isChecked());
					} else if (String.class.isAssignableFrom(fType)) {
						if (info.f.isAnnotationPresent(NotNull.class)) {
							if (((TextView) editor).getText().toString().trim().isEmpty())
								return false;
						}
						info.f.set(info.o, ((TextView) editor).getText().toString());
					} else if (editor instanceof ViewGroup)
						store((ViewGroup) editor);
				} catch (Exception e) {
					Log.e("RSCore", "Error while reading " + info.o.getClass().getName() + "." + info.f.getName()
							+ " from " + editor, e);
					return false;
				}
			}
		}
		return true;
	}

	private static Method findMethod(Object o, String name) {
		Class<?> c = o.getClass();
		while (c != Object.class) {
			try {
				Method m = c.getDeclaredMethod(name,View.class);
				return m;
			} catch (NoSuchMethodException nsm) {
			}
			c = c.getSuperclass();
		}
		return null;
	}

	private static class EditorInfo {
		private Object o;
		private Field f;

		public EditorInfo(Object o, Field f) {
			this.o = o;
			this.f = f;
		}
	}

	public static void bind(final Object o, final ViewGroup container, int intent) {
		if (o == null)
			return;

		Class<?> c = o.getClass();
		Context ctx = container.getContext();
		@SuppressWarnings("serial")
		List<Field> FIELDS = new ArrayList<Field>() {
			public boolean add(Field e) {
				int i = 0;
				int order = e.getAnnotation(SettingsField.class).order();
				boolean found = false;
				while (!found && (i < size())) {
					found = order < get(i).getAnnotation(SettingsField.class).order();
					if (!found)
						i++;
				}
				add(i, e);
				return true;
			};
		};

		while (c != Object.class) {
			FIELDS.clear();
			for (Field f : c.getDeclaredFields())
				if (f.isAnnotationPresent(SettingsField.class))
					FIELDS.add(f);
			for (Field f : FIELDS)
				try {
					SettingsField annotaion = f.getAnnotation(SettingsField.class);
					if (annotaion.hidden())
						continue;
					JSONObject options = new JSONObject(annotaion.options().isEmpty() ? " {}" : annotaion.options());
					f.setAccessible(true);
					LinearLayout ll = new LinearLayout(ctx);
					ll.setPadding(6 + intent, 2, 6, 2);
					ll.setOrientation(LinearLayout.VERTICAL);
					LinearLayout.LayoutParams lp = new LinearLayout.LayoutParams(LinearLayout.LayoutParams.MATCH_PARENT,
							LinearLayout.LayoutParams.WRAP_CONTENT);

					TextView label = new TextView(ctx);

					ll.setTag(new EditorInfo(o, f));
					Class<?> fType = f.getType();
					View editor = null;
					if (String.class.isAssignableFrom(fType)) {
						editor = new EditText(ctx);
						if (options.has("pick")) {
							final Method m = findMethod(o, options.getString("pick"));
							if (m != null) {
								m.setAccessible(true);
								((EditText) editor).setCompoundDrawablesRelativeWithIntrinsicBounds(0, 0,
										R.drawable.browse, 0);
								((EditText) editor).setOnTouchListener(new View.OnTouchListener() {
									@Override
									public boolean onTouch(View v, MotionEvent event) {
										if (event.getAction() == MotionEvent.ACTION_UP) {
											if (event.getRawX() >= (v.getRight()
													- ((TextView) v).getCompoundDrawables()[2].getBounds().width())) {
												try {
													m.invoke(o,v);
												} catch (InvocationTargetException | IllegalAccessException e) {

												}
												return true;
											}
										}
										return false;
									}
								});
							}
						}
						((EditText) editor).setText((String) f.get(o));
						if (!annotaion.type().isEmpty()) {
							if ("password".equals(annotaion.type()))
								((EditText) editor).setInputType(
										InputType.TYPE_CLASS_TEXT | InputType.TYPE_TEXT_VARIATION_PASSWORD);
							else if ("pin".equals(annotaion.type()))
								((EditText) editor).setInputType(
										InputType.TYPE_TEXT_VARIATION_PASSWORD | InputType.TYPE_CLASS_NUMBER);
							else if ("uri".equals(annotaion.type()))
								((EditText) editor)
										.setInputType(InputType.TYPE_CLASS_TEXT | InputType.TYPE_TEXT_VARIATION_URI);
							else if ("email".equals(annotaion.type()))
								((EditText) editor).setInputType(
										InputType.TYPE_CLASS_TEXT | InputType.TYPE_TEXT_VARIATION_EMAIL_ADDRESS);
						}
					} else if (Boolean.class == fType || boolean.class == fType) {
						label = new CheckBox(ctx);
						((CheckBox) label).setChecked(f.getBoolean(o));
					} else if (Integer.class == fType || int.class == fType) {
						if (options.has("type") && "list".equals(options.getString("type"))) {
							editor = new Spinner(ctx);
							SelectionAdapter a = new SelectionAdapter(ctx, options.getJSONArray("value"));
							((Spinner) editor).setAdapter(a);
							((Spinner) editor).setSelection(a.positionById(f.getInt(o)));
						} else {
							editor = new EditText(ctx);
							((EditText) editor).setText(String.valueOf(f.getInt(o)));
							((EditText) editor).setGravity(Gravity.END);
							((EditText) editor).setInputType(InputType.TYPE_CLASS_NUMBER);
						}
					} else if (Object.class.isAssignableFrom(fType)) {
						editor = new LinearLayout(ctx);
						((LinearLayout) editor).setOrientation(LinearLayout.VERTICAL);
						bind(f.get(o), (ViewGroup) editor, intent + 4);
						editor.setVisibility(View.GONE);
						label.setBackgroundColor(0xff3aff57);
						final View holder = editor;
						label.setOnClickListener(new View.OnClickListener() {
							@Override
							public void onClick(View v) {
								if (holder.getVisibility() == View.VISIBLE)
									holder.setVisibility(View.GONE);
								else
									holder.setVisibility(View.VISIBLE);
							}
						});
					}
					if (annotaion.descrId() != 0)
						label.setText(annotaion.descrId());
					else if (annotaion.description().isEmpty())
						label.setText(annotaion.name());
					else
						label.setText(annotaion.description());
					label.setPadding(6, 6, 6, 6);
					ll.addView(label, lp);
					if (editor != null)
						ll.addView(editor, lp);
					/*
					 * View v = new View(ctx);
					 * v.setBackgroundResource(android.R.drawable.divider_horizontal_dark);
					 * lp.height = 1; ll.addView(v,lp);
					 */
					container.addView(ll);
				} catch (Exception e) {
					Log.e("RSCore", "Build view ", e);
				}
			c = c.getSuperclass();

		}
	}

}
